import time

def retry(func, *args, tries=3, delay=0.0, exceptions=(Exception,), **kwargs):
    """
    Documentation for retry
    
    :param func: The function to be called and retried
    :param args: Any positional arguments for the function
    :param tries: The number of times the function to be tried
    :param delay: The delay between retries (in seconds)
    :param exceptions: The specific exceptions that trigger a retry
    :param kwargs: Any keyword arguments for the function

    This function returns the result of the function if successful. If not, it will raise the last exception it encountered.
    """
    if tries < 1:
        raise ValueError("Error: tries must be at least 1!")

    for attempt in range(tries):
        try:
            return func(*args, **kwargs)
        except exceptions as exception:
            if attempt == tries - 1:
                raise
            time.sleep(delay)

def retry_infinite(func, *args, delay=0.0, exceptions=(Exception,), **kwargs):
    """
    Documentation for retry_infinite
    
    :param func: The function to be called and retried
    :param args: Any arguments for the function
    :param delay: The delay between retries (in seconds)
    :param exceptions: The specific exceptions that trigger a retry
    :param kwargs: Any keyword arguments for the function

    This function returns the result of the function if successful. If not, it will repeat infinitely until it succeeds.
    """
    while True:
        try:
            return func(*args, **kwargs)
        except exceptions as exception:
            time.sleep(delay)


import pytest
from tiny_retry import retry, retry_infinite


def test_retry_succeeds_after_failures():
    state = {"n": 0}

    def flaky():
        state["n"] += 1
        if state["n"] < 3:
            raise ValueError("fail")
        return "ok"

    result = retry(flaky, tries=5, delay=0.0, exceptions=(ValueError,))
    assert result == "ok"
    assert state["n"] == 3  # it should stop retrying once it succeeds


def test_retry_raises_after_all_tries():
    state = {"n": 0}

    def always_fails():
        state["n"] += 1
        raise RuntimeError("nope")

    with pytest.raises(RuntimeError):
        retry(always_fails, tries=4, delay=0.0, exceptions=(RuntimeError,))

    assert state["n"] == 4  # should try exactly `tries` times


def test_retry_does_not_catch_unlisted_exception():
    state = {"n": 0}

    def raises_type_error():
        state["n"] += 1
        raise TypeError("wrong type")

    # We only retry ValueError, so TypeError should immediately escape
    with pytest.raises(TypeError):
        retry(raises_type_error, tries=5, delay=0.0, exceptions=(ValueError,))

    assert state["n"] == 1  # no retries happened


def test_retry_passes_args_and_kwargs():
    def add(a, b, scale=1):
        return (a + b) * scale

    assert retry(add, 2, 3, scale=10, tries=2) == 50


def test_retry_rejects_invalid_tries():
    with pytest.raises(ValueError):
        retry(lambda: 123, tries=0)


def test_retry_infinite_eventually_succeeds():
    state = {"n": 0}

    def flaky():
        state["n"] += 1
        if state["n"] < 3:
            raise ValueError("try again")
        return "done"

    assert retry_infinite(flaky, delay=0.0, exceptions=(ValueError,)) == "done"
    assert state["n"] == 3
